import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.optim
from torch.autograd import Variable
import os
import numpy as np
from PIL import Image
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

import configparser
from util import Recoder, paramNumber, kspace_subsampling_pytorch, imgFromSubF_pytorch

from data_process import getDataloader
import random
from network.cascadeNetwork import CN_Dense

class core():
    def __init__(self, configPath, single_image_test=False):
        self.config = configparser.ConfigParser()
        self.config.read(configPath)

        self.batchSize = int(self.config['train']['batchSize'])
        self.epoch = int(self.config['train']['epoch'])
        self.LR = float(self.config['train']['learningRate'])
        
        self.saveEpoch = int(self.config['log']['saveEpoch'])
        self.maxSaved = int(self.config['log']['maxSaved'])

        self.useCuda = self.config.getboolean('general','useCuda')
        self.needParallel = self.config.getboolean('general','needParallel')
        if(self.useCuda):
            self.dtype = torch.cuda.FloatTensor
        else:
            self.dtype = torch.FloatTensor


        print("#Result Path:"+self.config['general']['path'])
        print("#Create network:"+self.config['general']['netType'])

        if self.useCuda:
            listDeviceStr = self.config['general']['device']
            listDevice = []
            for i in range(len(listDeviceStr)):
                if(listDeviceStr[i] == '1'):
                    listDevice.append(str(i))
                devicesStr=','.join(listDevice)
            os.environ["CUDA_VISIBLE_DEVICES"] = devicesStr
            assert len(listDevice) > 0, "No device is selected"

        self.net = CN_Dense(2, c=5, dilate=True, useOri=True, transition=0.5, trick=2).type(self.dtype)

        if(self.needParallel):
            print('parallel device:'+str(listDevice))
            self.net = nn.DataParallel(self.net, device_ids = list(range(len(listDevice)))).type(self.dtype)
        self.lossForward = torch.nn.MSELoss().type(self.dtype)
        paramNum = paramNumber(self.net)
        print("#Number of network parameters:%d"%paramNum)
        if not single_image_test:
            if self.config.getboolean('general', 'use_cal'):
                cal = float(self.config['general']['cal'])
            else:
                cal = None
            self.trainloader, self.trainsetSize = getDataloader(self.config['general']['dataType'], 'train', self.batchSize, us_rate=float(self.config['general']['us_rate']), is_train=True, imgpath=self.config['general']['train_path'], cal=cal)
            self.testloader, self.testsetSize = getDataloader(self.config['general']['dataType'], 'test', 1, us_rate=float(self.config['general']['us_rate']), is_train=False, imgpath=self.config['general']['vali_path'], cal=cal)
        
        self.record = Recoder(self.config['general']['path'], self.saveEpoch*self.maxSaved)
        

        self.config.write(open(self.record.rootPath+"/config.ini","w"))
        self.record.logNet(self.net)
        self.record.log("### Number of network parameters:%d"%paramNum)


        self.weightDecay = float(self.config['train']['weightdecay'])

        print('#Optimizer: '+self.config['train']['optimizer']+' LR = %.2e weightDecay = %.2e'%(self.LR, self.weightDecay))
        self.ckp_epoch = 0

    def Separate_test(self, load_path):
        self.net.load_state_dict(torch.load(load_path))
        l, p1, p2, s1, s2 = self.validation()
        return p1, p2, s1, s2


    def train(self, epoch = 0, need_evaluate = True):
        if(epoch == 0):
            epoch = self.epoch
        self.optimizer = torch.optim.Adam(self.net.parameters(), lr=self.LR, weight_decay = self.weightDecay)
        msg = "start training: epoch = %d"%(epoch)
        self.record.log(msg)
        print(msg)
        for j in range(1,epoch + 1):
            self.net.train()
            i = 0
            total_loss = 0
            for mode, label, mask in self.trainloader:
                self.optimizer.zero_grad()
                #mask = torch.randint(2, [8, 256, 256])
                netLabel = Variable(label).type(self.dtype)
                mask_var = Variable(mask).type(self.dtype)
                subF = kspace_subsampling_pytorch(netLabel,mask_var)
                complexFlag = (mode[0] == 'complex')
                netInput = imgFromSubF_pytorch(subF,complexFlag)
                netOutput = self.net(netInput, subF, mask_var)
                loss = self.lossForward(netOutput,netLabel)
                loss.backward()
                total_loss = total_loss + loss.item()*label.shape[0]

                self.optimizer.step()
                i += label.shape[0]
                if i % (100) == 0:
                    print('Epoch %05d [%04d/%04d] loss %.8f' % (j+self.ckp_epoch,i,self.trainsetSize,loss.item()))
            self.record.log_train(j+self.ckp_epoch,total_loss/i)

            if j % self.saveEpoch == 0:
                print('Epoch %05d [%04d/%04d] loss %.8f SAVED' % (j+self.ckp_epoch,i,self.trainsetSize,total_loss/self.trainsetSize))
                if need_evaluate:
                    l,p1,p2,s1,s2 = self.validation()
                    self.record.log_valid(j+self.ckp_epoch,l,p1,p2,s1,s2)
                    self.record.log("Evaluate psnr(before|after) =  %.2f|%.2f ssim = %.4f|%.4f"%(p1,p2,s1,s2))
                self.record.write_to_file(self.net.state_dict(), False)
        self.ckp_epoch += epoch
        self.record.write_to_file(self.net.state_dict(), True)
    
    def testValue(self,mode,label,mask):
        result = self.test(mode,label,mask)
        
        return result['loss'],result['psnr1'],result['psnr2'],result['ssim1'],result['ssim2']
    
    def test(self,mode,label,mask):
        y = label.numpy() 
        netLabel = Variable(label).type(self.dtype)

        mask_var = Variable(mask).type(self.dtype)
        subF = kspace_subsampling_pytorch(netLabel,mask_var)
        complexFlag = (mode[0] == 'complex')
        netInput = imgFromSubF_pytorch(subF,complexFlag)
            
        y=y[0]
            
        netOutput = self.net(netInput, subF, mask_var)
        loss = self.lossForward(netOutput,netLabel)
        netOutput_np = netOutput.cpu().data.numpy()
        img1 = netOutput_np[0,0:1].astype('float64')
        if(netOutput_np.shape[1]==2):
            netOutput_np = abs(netOutput_np[:,0:1]+netOutput_np[:,1:2]*1j)
        img2 = netOutput_np[0].astype('float64')
        y2 = y[0:1]
    
        img1 = np.clip(img1,0,1)
        img2 = np.clip(img2,0,1)
        
        if(self.isFastMRI):
            psnrBefore = psnr(y2,img1,12)
            psnrAfter = psnr(y2,img2,12)
        else:
            psnrBefore = psnr(y2,img1)
            psnrAfter = psnr(y2,img2)
        
        ssimBefore = ssim(y2[0],img1[0])
        ssimAfter = ssim(y2[0],img2[0])
        
        return {"loss":loss.item(),"psnr1":psnrBefore,"psnr2":psnrAfter,"ssim1":ssimBefore,"ssim2":ssimAfter,"result1":img1,"result2":img2,'label':y2}

    def single_test(self, mode, label, mask):
        y = label.numpy()
        netLabel = Variable(label).type(self.dtype)

        mask_var = Variable(mask).type(self.dtype)
        subF = kspace_subsampling_pytorch(netLabel, mask_var)
        complexFlag = (mode == 'complex')
        netInput = imgFromSubF_pytorch(subF, complexFlag)

        y = y[0]

        netOutput = self.net(netInput, subF, mask_var)
        loss = self.lossForward(netOutput, netLabel)
        netOutput_np = netOutput.cpu().data.numpy()
        img1 = netOutput_np[0, 0:1].astype('float64')
        if (netOutput_np.shape[1] == 2):
            netOutput_np = abs(netOutput_np[:, 0:1] + netOutput_np[:, 1:2] * 1j)
        img2 = netOutput_np[0].astype('float64')
        y2 = y[0:1]

        img1 = np.clip(img1, 0, 1)
        img2 = np.clip(img2, 0, 1)


        psnrBefore = psnr(y2, img1)
        psnrAfter = psnr(y2, img2)

        ssimBefore = ssim(y2[0], img1[0])
        ssimAfter = ssim(y2[0], img2[0])

        return {"loss": loss.item(), "psnr1": psnrBefore, "psnr2": psnrAfter, "ssim1": ssimBefore, "ssim2": ssimAfter,
                "result1": img1, "result2": img2, 'label': y2, 'rec_1': img1, 'rec_2': img2}


    def validation(self, returnList=False):
        self.net.eval()
        i = 0
        totalLoss = 0
        lpsnr1 = []
        lpsnr2 = []
        lssim1 = []
        lssim2 = []
        psnr1 = 0
        psnr2 = 0
        ssim1 = 0
        ssim2 = 0
        for mode, label, mask in self.testloader:
            loss0, psnrA, psnrB, ssimA, ssimB = self.testValue(mode, label, mask)

            totalLoss += loss0
            psnr1 += psnrA
            psnr2 += psnrB
            ssim1 += ssimA
            ssim2 += ssimB
            lpsnr1.append(psnrA)
            lpsnr2.append(psnrB)
            lssim1.append(ssimA)
            lssim2.append(ssimB)
            i += 1
            print('Evaluating %04d psnr(before|after) =  %.2f|%.2f ssim = %.4f|%.4f' % (i, psnrA, psnrB, ssimA, ssimB),
                  '\r', end='')
        print('Evaluating %04d psnr(before|after) =  %.2f|%.2f ssim = %.4f|%.4f Done' % (
        i, psnr1 / i, psnr2 / i, ssim1 / i, ssim2 / i))
        if (returnList):
            return lpsnr1, lpsnr2, lssim1, lssim2
        return totalLoss / i, psnr1 / i, psnr2 / i, ssim1 / i, ssim2 / i

    def single_validation(self, mode, imagename, load_path, us_rate):
        self.net.load_state_dict(torch.load(load_path))
        self.net.eval()
        im = Image.open(imagename)
        im = im.resize((256, 256), Image.BILINEAR)
        im_np = np.array(im).astype(np.float32) / 255.
        if len(im_np.shape) == 3:
            im_np = im_np[:,:,0]
        if mode == 'complex':
            label = np.zeros((2, 256, 256))
            label[0] = im_np
        else:
            label = im_np
        pickindex = random.sample(np.arange(0, label.shape[-2] * label.shape[-1]).tolist(),
                                  np.int64(us_rate * label.shape[-2] * label.shape[-1]))
        mask = np.zeros((label.shape[-2] * label.shape[-1],))
        mask[pickindex] = 1
        mask = np.reshape(mask, (label.shape[-2], label.shape[-1]))
        mask = np.fft.ifftshift(mask)
        results = self.single_test(mode,  torch.tensor(label).unsqueeze(0), torch.tensor(mask).cuda().unsqueeze(0))

        return results
    
    def loadCkpt(self, expectedEpoch, isChecked=False):
        self.record.load_from_file(self.net, expectedEpoch, isChecked, False)
        self.ckp_epoch = expectedEpoch

